#ifndef ACTIONSCHEDULER_H_
#define ACTIONSCHEDULER_H_

#include "Connection.h"
#include "WorldModel.h"
#include "Logger.h"

#include <queue>

extern Logger Log;

using namespace std;

/**
 * Used for callback parameter storage
 */ 
struct scheduled_action {
	/**
	 * Indicates cycle, where action shall be performed
	 */
	int cycle;
	
	/**
	 * Contains method, that should be executed on given cycle
	 */ 
	void (*action_callback)(WorldModel*,ServerSettings*,PlayerSettings*, void*, int);
	
	/**
	 * Contains parameter reference, that should be passed to action method
	 */ 
	void* params;
	
	/**
	 * Size of parameter that is passed using pointer
	 */ 
	int params_size;
	
	/**
	 * Method for cleanup of parameters
	 */
	void (*cleanup_callback)(void*, int);
};

/**
 * This component is responsible for execution of registered actions in specific time.
 * 
 * Reason for creation of this component is need reinforced learning. Reinforced learning need
 * feedback on perfomed action. In robocup environment action effect can't be evaluated ad once.
 * This component is used for later evaluation of state space and update of method that is used to 
 * choose actions from action space.
 */
class ActionScheduler
{
	WorldModel     *WM;        /*!< Worldmodel containing all data of the match*/
	ServerSettings *SS;        /*!< ServerSettings with all server settings    */
	PlayerSettings *PS;        /*!< PlayerSettings with all client settings    */
	
	queue<struct scheduled_action> scheduled_actions;
protected:
	void doDispatchAction(struct scheduled_action *action);
	
public:
	ActionScheduler(WorldModel* wm, ServerSettings *ss, PlayerSettings *ps);
	virtual ~ActionScheduler();
	
	/**
	 * Executes actions that are scheduled to specific cycle. When actions for this cycle are 
	 * executed, they are discarted.
	 * @param cycle Indicates cycle where actions needs to be performed
	 */
	void dispatchAction(int cycle);
	
	
	void scheduleAction(int cycle, void (*action_callback)(WorldModel*,ServerSettings*,PlayerSettings*, void*, int), 
			void (*cleanup_callback)(void*, int), void* params, int params_size); 
};

#endif /*ACTIONSCHEDULER_H_*/
