/*
Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Amsterdam nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*! \file PlayerTeams.cpp
<pre>
<b>File:</b>          PlayerTest.cpp
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       10/12/2000
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the class definitions for the
                      Player that are used to test the teams' high level
                      strategy.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
10/12/2000        Jelle Kok       Initial version created
</pre>
*/

#include "Player.h"
#include "ArsToolbox.h"
#include "SoccerTypes.h"
#include <Math.h>
#include <stdio.h>


#define LOG_SITUATIONS


SoccerCommand Player::kickToScore(  ){
    SoccerCommand soc(CMD_ILLEGAL);
  	//////////////////////////////////////////////////////////////////////////
	// FUZZY REGULACIA	START												//
	//////////////////////////////////////////////////////////////////////////

	double goalWidth = SS->getGoalWidth();
	double a = goalWidth * 0.5;
	int samples = 100;

	const int IN_Y_GOALIE = 0;			// input  fuzzygroup (lingvisticka premenna)
	const int IN_ME_TO_GOALIE = 1;		// input  fuzzygroup (lingvisticka premenna)
	const int OUT_Y_DIRECTION = 0;		// output fuzzygroup (lingvisticka premenna)

	const int BRANKAR_VLAVO = 0;		// charakteristicka funkcia pre IN_Y_GOALIE
	const int BRANKAR_TROCHU_VLAVO = 1; // charakteristicka funkcia pre IN_Y_GOALIE
 	const int BRANKAR_TROCHU_VPRAVO = 2;// charakteristicka funkcia pre IN_Y_GOALIE
	const int BRANKAR_VPRAVO = 3;		// charakteristicka funkcia pre IN_Y_GOALIE

	const int NALAVO_OD_BRANKARA  = 0; 	// charakteristicka funkcia pre IN_ME_TO_GOALIE
	const int BRANKAR_PREDOMNOU   = 1; 	// charakteristicka funkcia pre IN_ME_TO_GOALIE
	const int NAPRAVO_OD_BRANKARA = 2;	// charakteristicka funkcia pre IN_ME_TO_GOALIE

	const int KOPNI_EX_VLAVO = 0;		// charakteristicka funkcia pre OUT_Y_DIRECTION
	const int KOPNI_VLAVO = 1;			// charakteristicka funkcia pre OUT_Y_DIRECTION
	const int KOPNI_VPRAVO = 2;			// charakteristicka funkcia pre OUT_Y_DIRECTION
	const int KOPNI_EX_VPRAVO = 3;		// charakteristicka funckia pre OUT_Y_DIRECTION

	cFuzzyObj * fz = new cFuzzyObj();
 	fz->SetSamples(samples);

	fz->AddInput(-a, a);	// vytvorenie IN_Y_GOALIE
 	fz->AddInput(-a, a);	// vytovrenie IN_ME_TO_GOALIE
  	fz->AddOutput(-a, a);	// vytvorenie OUT_Y_DIRECTION

  	fz->AddInputMF(IN_Y_GOALIE, FZ_LEFT,      -0.75*a,  -0.4*a);					// BRANKAR_VLAVO
 	fz->AddInputMF(IN_Y_GOALIE, FZ_TRAPEZOID,  -0.6*a,  -0.3*a, -0.03*a, 0.0*a); 	// BRANKAR_TROCHU_VLAVO
   	fz->AddInputMF(IN_Y_GOALIE, FZ_TRAPEZOID,  0.0*a,   0.03*a,  0.3*a, 0.6*a);		// BRANKAR_TROCHU_VPRAVO
 	fz->AddInputMF(IN_Y_GOALIE, FZ_RIGHT, 	    0.4*a,   0.75*a);					// BRANKAR_VPRAVO

	fz->AddInputMF(IN_ME_TO_GOALIE, FZ_LEFT,   	  -0.9*a, -0.2*a);					// NALAVO_OD_BRANKARA
	fz->AddInputMF(IN_ME_TO_GOALIE, FZ_TRAPEZOID, -0.3*a, -0.2*a, 0.2*a, 0.3*a);	// BRANKAR_PREDOMNOU
	fz->AddInputMF(IN_ME_TO_GOALIE, FZ_RIGHT,      0.2*a,  0.9*a);					// NAPRAVO_OD_BRANKARA


	fz->AddOutputMF(OUT_Y_DIRECTION, FZ_TRIANGLE,      -a, -0.9*a, -0.8*a);			// KOPNI_EX_VLAVO
	fz->AddOutputMF(OUT_Y_DIRECTION, FZ_TRAPEZOID, -0.9*a, -0.8*a, -0.6*a, -0.4*a);	// KOPNI_VLAVO
	fz->AddOutputMF(OUT_Y_DIRECTION, FZ_TRAPEZOID,  0.4*a,  0.6*a,  0.8*a,  0.9*a); // KOPNI_VPRAVO
	fz->AddOutputMF(OUT_Y_DIRECTION, FZ_TRIANGLE,   0.8*a,  0.9*a, a);				// KOPNI_EX_VPRAVO

	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_VLAVO, NALAVO_OD_BRANKARA , KOPNI_EX_VLAVO);
	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_VLAVO, BRANKAR_PREDOMNOU  , KOPNI_VPRAVO);
	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_VLAVO, NAPRAVO_OD_BRANKARA, KOPNI_VPRAVO);

	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_TROCHU_VLAVO, NALAVO_OD_BRANKARA  , KOPNI_EX_VLAVO);
	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_TROCHU_VLAVO, BRANKAR_PREDOMNOU   , KOPNI_EX_VPRAVO);
	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_TROCHU_VLAVO, NAPRAVO_OD_BRANKARA , KOPNI_VPRAVO);

  	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_TROCHU_VPRAVO, NALAVO_OD_BRANKARA  , KOPNI_VLAVO);
 	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_TROCHU_VPRAVO, BRANKAR_PREDOMNOU   , KOPNI_EX_VLAVO);
 	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_TROCHU_VPRAVO, NAPRAVO_OD_BRANKARA , KOPNI_EX_VPRAVO);

  	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_VPRAVO, NALAVO_OD_BRANKARA  , KOPNI_VLAVO);
  	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_VPRAVO, BRANKAR_PREDOMNOU   , KOPNI_VLAVO);
  	fz->AddRule(OUT_Y_DIRECTION, BRANKAR_VPRAVO, NAPRAVO_OD_BRANKARA , KOPNI_EX_VPRAVO);


	VecPosition gp = WM->getGlobalPosition(OBJECT_OPPONENT_GOALIE);
	double opp_goalie_position = gp.getY();
 	if(opp_goalie_position < -a) {
  		opp_goalie_position = -a;
  	}
  	if(opp_goalie_position > a) {
   		opp_goalie_position = a;
   	}

 	VecPosition myPos = WM->getAgentGlobalPosition();
	VecPosition diff  = myPos - gp;
  	double my_rel_pos     = diff.getY();

	if(my_rel_pos<0) {
		if(my_rel_pos < -a) {
			my_rel_pos = -a;
 		}
 	}
 	else {
	if(my_rel_pos > a) {
		my_rel_pos = a;
	}
	}

 	double y_direction = fz->CountResult(OUT_Y_DIRECTION, DFZ_CENTROID, OPS_ZADEH, opp_goalie_position, my_rel_pos);
 	delete fz;

 	//////////////////////////////////////////////////////////////////////////
 	// FUZZY REGULACIA	END													//
  	//////////////////////////////////////////////////////////////////////////

	VecPosition posGoal( PITCH_LENGTH/2.0, y_direction );
	soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal

  	ACT->putCommandInQueue( soc );
  	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  	Log.log( 100, "kick ball" );
  	return soc;
}

SoccerCommand Player::kickToDrible(  ) {
    SoccerCommand soc(CMD_ILLEGAL);
    Log.log(100, "Kicking to drible");
	// TODO: find better position to kick ball to, now random
  	VecPosition posGoal( PITCH_LENGTH/2.0,
          (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth() );
  	soc = kickTo( posGoal, SS->getBallSpeedMax() / 30 ); // kick maximal

  	ACT->putCommandInQueue( soc );
  	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  	Log.log( 100, "kick ball" );
}


/** GOS - passTo fitness function.
 * 
 * Determines fitness function for passing ball from one position to another.
 * 
 * @param from1X line start X
 * @param from1Y line start Y
 * @param to1X line end X
 * @param to1Y line end Y
 * @param pX player X
 * @param pY player Y
 * @param toX pass ball to X
 * @param toY pass ball to Y
 */
double Player::passToFitness(double from1X, double from1Y, double to1X, double to1Y, double pX, double pY, double toX, double toY, double r) {
    //if (r == -1)
    //    r = getInterestRadius(VecPosition::pointsDistance(from1X, from1Y, toX, toY));

    double dist = VecPosition::pointsDistance(from1X, from1Y, toX, toY);        
    double fdist1 = (30 * dist / RAY_LENGTH); 
    double fdist2 = (ArsToolbox::linePointDistance(
             from1X,
             from1Y,
             to1X,
             to1Y,
             pX,
             pY
        ) / (r + 1.0) + 1.0);          
    double f = fdist1 / fdist2;
    if (dist > 20)
        f -= 2 * (RAY_LENGTH - dist) / RAY_LENGTH; 
    
    return ArsToolbox::trim(f, 0, 12);
}

/** GOS - passTo fitness function for teammate.
 * Determines fitness function for passing ball from one position to anothers.
 * anothers.
 * 
 * @param from1X line start X
 * @param from1Y line start Y
 * @param to1X line end X
 * @param to1Y line end Y
 * @param pX player X
 * @param pY player Y
 * @param toX pass ball to X
 * @param toY pass ball to Y
 */
double Player::passToTeammateFitness(double from1X, double from1Y, double to1X, double to1Y, double pX, double pY, double toX, double toY, double r) {
    double f1 = ArsToolbox::trim( 
        1.3 * passToFitness(
        from1X,
        from1Y,
        to1X,
        to1Y,
        pX,
        pY,
        toX,
        toY,
        r
    ), 0, 12);
    
    // backward/forward correction
    double f2 =
        (toX - from1X >= 0 ? 0.5 : -1.5)
            +             
        ArsToolbox::trim(
            2 * (toX - from1X)
                / 
            ArsToolbox::trim(
                PITCH_LENGTH - goalMiddle.getDistanceTo(toX, toY),
                0.1,
                PITCH_LENGTH),
            -4,
            2
        );
    
    return ArsToolbox::trim(f1 + f2, 0, 12);
}

/** GOS - passTo fitness function for opponent
 * 
 * Determines fitness function for passing ball from one position to anothers.
 * 
 * @param from1X line start X
 * @param from1Y line start Y
 * @param to1X line end X
 * @param to1Y line end Y
 * @param pX player X
 * @param pY player Y
 * @param toX pass ball to X
 * @param toY pass ball to Y
 */
double Player::passToOpponentFitness(double from1X, double from1Y, double to1X, double to1Y, double pX, double pY, double toX, double toY, double r) {
    return 0.7 * passToFitness(
        from1X,
        from1Y,
        to1X,
        to1Y,
        pX,
        pY,
        toX,
        toY,
        r
    ) + 3.0;          
}

bool Player::isPassBlocked(ObjectT oPlayer, AngRad ang) {
}

double Player::leadingPass(ObjectT oPlayer, AngRad ang) {
    
}

/** GOS - Determinse fitness for specific angle and determines the best position.
 * 
 * Using circle algorithm this function finds the best position on 
 * line specified by angle
 * Here is made the decision if a player (teammate or oponent) will be 
 * considered.
 * 
 * @see passToFitness
 * 
 * @param angle line angle from player
 * @param oPlayer player to kick to, if function should find the best
 *                then use OBJECT_SELECT_PLAYER
 * @param pos the best position where to kick ball to
 * @param dConfThr confidence treshold for players
 * @result fitness for this angle
 */
double Player::passTo(  AngRad angle, ObjectT oPlayer, VecPosition 
                        *pos, double dConfThr) {
    VecPosition vec1 = WM->getAgentGlobalPosition();    // agent position
    double x, y;                                        // line is to
    double nx, ny, nd;                                  // pass to
    double fit, tfit = -12, ofit = 0;                   // fitness
    int iIndex;                                         // indexer
    double p1x, p1y, p2x, p2y;                          // crosspoints
    VecPosition posObj(0, 0);                                 // players pos
    double dDist;                                       // distance
    double r;

    // added by petras
    // increase the confidence threshold for goalkeeper
    if( formations->getPlayerType()== PT_GOALKEEPER )
    	dConfThr = 0.5;    
    
    // create line at this angle in 
    ArsToolbox::createLineTo(vec1.getX(), vec1.getY(), angle, RAY_LENGTH, &x, &y);
    nx = 0;
    ny = 0;
    nd = RAY_LENGTH;
    
    // find teammates  
    for( ObjectT o = WM->iterateObjectStart( iIndex, OBJECT_SET_TEAMMATES_NO_GOALIE, dConfThr);
         o != OBJECT_ILLEGAL;
         o = WM->iterateObjectNext ( iIndex, OBJECT_SET_TEAMMATES_NO_GOALIE, dConfThr) )
    {
        // get player position and his distance
        posObj    = WM->getGlobalPosition( o );
        dDist     = posObj.getDistanceTo( vec1 );
        
        // find interest radius of circle for this player
        r         = getInterestRadius(dDist);
      
        // try to find intersection with player
        if (dDist > 1 && ArsToolbox::circleLineIntersection(
            vec1.getX(), 
            vec1.getY(), 
            x, 
            y, 
            posObj.getX(), 
            posObj.getY(),
            r, 
            &p1x, 
            &p1y, 
            &p2x, 
            &p2y) > 0) 
        {
            // calculate teammate fitness            
            fit = passToTeammateFitness(
                    vec1.getX(),
                    vec1.getY(),
                    x,
                    y,
                    posObj.getX(), 
                    posObj.getY(),
                    p2x,
                    p2y,
                    r
                );            
            // is this fitness better?
            if (fit > tfit) {
                tfit = fit;
                nx = p2x;
                ny = p2y;
                nd = dDist;
            }

#ifdef LOG_SITUATIONS                
            // log to file...
            log->addObject(LOG_PLAYER_TEAM, posObj.getX(),posObj.getY());
            log->addObject(LOG_CIRCLE, posObj.getX(), posObj.getY(), r);
#endif
        }      
    }
    WM->iterateObjectDone( iIndex );
      
    // find opponents
    for( ObjectT o = WM->iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS, dConfThr );
         o != OBJECT_ILLEGAL;
         o = WM->iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS, dConfThr ) )
    {
        posObj    = WM->getGlobalPosition( o );
        dDist     = posObj.getDistanceTo( vec1 );
        r         = getInterestRadius(dDist); 
        
        // try to find intersection with opponents which are before best player
        if ((dDist < nd + 8.0) && ArsToolbox::circleLineIntersection(
            vec1.getX(), 
            vec1.getY(), 
            x, 
            y, 
            posObj.getX(), 
            posObj.getX(),
            r, 
            &p1x, 
            &p1y, 
            &p2x, 
            &p2y) > 0) 
        {
            fit = passToOpponentFitness(
                 vec1.getX(),
                 vec1.getY(),
                 x,
                 y,
                 posObj.getX(), 
                 posObj.getY(),
                 p1x,
                 p1y,
                 r
            );            

            if (fit > ofit)
                ofit = fit;
#ifdef LOG_SITUATIONS
            // log situations
            log->addObject(LOG_PLAYER_OTHER, posObj.getX(), posObj.getY());
            log->addObject(LOG_CIRCLE, posObj.getX(), posObj.getY(), r);
#endif
        }      
    }
    WM->iterateObjectDone( iIndex );
     
    if (pos != NULL) {
        pos->setX(nx);
        pos->setY(ny);
    }
    
#ifdef LOG_SITUATIONS
    // log line
    log->addObject(LOG_PLAYER_OTHER, posObj.getX(), posObj.getY());
    log->addObject(LOG_LINE, vec1.getX(), vec1.getY(), nx, ny, "white", ArsToolbox::trim(tfit - ofit, 0, 12));
#endif

    return ArsToolbox::trim(tfit - ofit, 0, 12); 
}

/** GOS - Kick to pass with circle algorithm.
 * 
 * Finds the best ray to kick ball to using ray fitness function
 * Algorithm scans every N degree, where N = 2PI / KICK_RAYS_NR
 * 
 * @param dFitness output fitness of the command
 * 
 * @see passTo
 * @result Used and sent SoccerCommand 
 */
SoccerCommand Player::circleKickToPass(double* dFitness ) {
    VecPosition vec1 = WM->getAgentGlobalPosition();
    VecPosition pos(0, 0), tpos(0, 0);
    double tfit, fit = -12;
  
    // find the best ray and position to kick ball to
    for (int i = 0; i < KICK_RAYS_NR; i++) {
        tfit = passTo(
            2.0 * M_PI * (((double)i + 1) / (double)KICK_RAYS_NR),
            OBJECT_SELECT_PLAYER, 
            &tpos
        );
#ifdef LOG_SITUATIONS
        char msg[50];
        sprintf(msg, "UHOL: %.2g, FIT: %.4g", (180/M_PI) * 2.0 * M_PI * (((double)i + 1) / (double)KICK_RAYS_NR), tfit );
    log->addObject(LOG_LINE, vec1.getX(), vec1.getY(), pos.getX(), pos.getY(), msg);
#endif
        if (tfit > fit) {
            fit = tfit;
            pos.setX(tpos.getX());
            pos.setY(tpos.getY());
        }
    }

    // calculate ball speed to reech specified position
    double speed = ArsToolbox::trim(pos.getDistanceTo(vec1)/50 * SS->getBallSpeedMax(), 0, SS->getBallSpeedMax());
    
    // send this command
    SoccerCommand soc(CMD_ILLEGAL);
    soc = kickTo( pos, speed );    
#ifdef LOG_SITUATIONS
    // log position where ball will be kicked
    log->addObject(LOG_PLAYER_ACTIVE, vec1.getX(), vec1.getY());    
    log->addObject(LOG_LINE, vec1.getX(), vec1.getY(), pos.getX(), pos.getY(), "Yellow");
#endif

    if (dFitness != NULL)
        *dFitness = fit; 

    return soc;
}

/** GOS - Player passes ball to another player.
 * 
 * Position to kick to is determined with prediction
 * 
 * @param dFitness output fitness of the command
 * 
 */ 
SoccerCommand Player::kickToPass(double* dFitness){
    SoccerCommand soc(CMD_ILLEGAL);

#ifdef LOG_SITUATIONS
    log->situationBegin("Prihravka", WM->getCurrentCycle());
#endif  
    
    soc = circleKickToPass(dFitness);  
	
#ifdef LOG_SITUATIONS
    log->situationEnd();
#endif
    return soc;
}

/** GOS - method return a comand to kick the ball when OUT
 */
SoccerCommand Player::kickToPlay() {
	SoccerCommand soc(CMD_ILLEGAL);
  	double speed;  
  	double dis;	
  	ObjectT obj;
  
	obj = WM->getClosestRelativeInSet(OBJECT_SET_TEAMMATES_NO_GOALIE, &dis, 2.0, NULL);
		
	// ak napriek tomu nemame ziadneho hraca, tak kopeme na branku  	
	if (obj == OBJECT_ILLEGAL){
		kickToScore();
	} else {
		VecPosition vec1 = WM->getGlobalPosition(obj); 
		speed = SS->getBallSpeedMax() * 0.5;
		soc = kickTo(vec1, speed);
		ACT->putCommandInQueue( soc );
	  	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
	}
    return soc;	
} 

/*!This method is the first complete simple team and defines the actions taken
   by all the players on the field (excluding the goalie). It is based on the
   high-level actions taken by the simple team FC Portugal that it released in
   2000. The players do the following:
   - if ball is kickable
       kick ball to goal (random corner of goal)
   - else if i am fastest player to ball 
       intercept the ball
   - else
       move to strategic position based on your home position and pos ball */
SoccerCommand Player::deMeer5(  )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp;

	if( WM->isBeforeKickOff( ) )
	{
		if( WM->isKickOffUs( ) && WM->getPlayerNumber() == 9 ) // 9 takes kick
    	{
      		if( WM->isBallKickable() )
      		{      	
        		VecPosition posGoal( PITCH_LENGTH/2.0,
                             (-1 + 2*(WM->getCurrentCycle()%2)) * 
                             0.4 * SS->getGoalWidth() );
                            
        		soc = kickToPass(); //nahraj
        		Log.log( 100, "take kick off" );       		
       		}
      		else
      		{
        		soc = intercept( false );  
        		Log.log( 100, "move to ball to take kick-off" );
      		}  
            ACT->putCommandInQueue( soc );
            ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		return soc;
    	}  
    	if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        	posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )  
    	{
      		formations->setFormation( FT_INITIAL );       // go to kick_off formation
      		ACT->putCommandInQueue( soc=teleportToPos( WM->getStrategicPosition() ));
    	}
    	else                                            // else turn to center
    	{
      		ACT->putCommandInQueue( soc=turnBodyToPoint( VecPosition( 0, 0 ), 0 ) );
      		ACT->putCommandInQueue( alignNeckWithBody( ) );
    	}
  	}
  	// ak mame rozohrat, tak takto
  	else if(WM->isDeadBallUs() && !WM->isGoalKickUs()) {
  		if(WM->isBallKickable()) {
  			soc = kickToPlay();
  			//soc = kickToScore();
  		}
  		else if(WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp)
              == WM->getAgentObjectType()) {
              	
            //soc = intercept( false );                      // intercept the ball
            soc = moveToPos(posBall, PS->getPlayerWhenToTurnAngle());
      		if( soc.commandType == CMD_DASH &&             // if stamina low
          			WM->getAgentStamina().getStamina() <
             		SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      		{
        		soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        		ACT->putCommandInQueue( soc );
        		ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		}
      		else                                           // if stamina high
      		{
        		ACT->putCommandInQueue( soc );               // dash as intended
        		ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		}
      		return soc;
  		}
	  /*else if(WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr()){	// ak neviem, kde je lopta, tak ju idem hladat
  			ACT->putCommandInQueue( soc = searchBall() );   // najdi loptu
      		ACT->putCommandInQueue( alignNeckWithBody( ) ); // natoc sa
  		}*/
  		else {	// otacaj sa, aby si vsetkych videl
  			ACT->putCommandInQueue( soc = turnAround());
  			ACT->putCommandInQueue( alignNeckWithBody( ) ); // natoc sa
  		}
  	}
  	else if(WM->isDeadBallThem()) {
		ACT->putCommandInQueue( soc = turnAround());
  		ACT->putCommandInQueue( alignNeckWithBody( ) ); // natoc sa
  	}
  	else
  	{
  		
  		 				/*	Jahodovi Princovia, 2007/2008
  		  		       * 
  		  		       * 	This method checks the actual ball position and according to the value
  		  		       * 	changes formation
  		  		       */      
  		  		      
  		  				//if ball is on the opponent's half
  		  		      if(WM->getBallPos().getX() > 0)
  		  		    	  formations->setFormation(FT_334_OFFENSIVE);
  		  		      else 
  		  		    	formations->setFormation(FT_433_OFFENSIVE);
  		  		      
    	
    	soc.commandType = CMD_ILLEGAL;
    	
    	double distance = posAgent.getDistanceTo(posBall);
    
    	// nastavenie pohladu
    	SoccerCommand sc = adjustView(distance);
    	if(sc.commandType != CMD_ILLEGAL) {
    		ACT->putCommandInQueue(sc);
    	}

    	if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    	{
      		ACT->putCommandInQueue( soc = searchBall() );   // if ball pos unknown
      		ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    	}
    	else if( WM->isBallKickable())                    // if kickable
    	{
    		// DRIBBLE
    
            // if i can dribble or i am not near to goal
      		if( canDribbleWithBall() != 0 && 
      		    posAgent.getDistanceTo(goalMiddle) >= (PITCH_LENGTH / 7.5) )
      		{
      			soc = handleWithBall();
      			ACT->putCommandInQueue( soc );
      	    	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		}
    		// KONIEC DRIBLOVANIE
      		else
      		{
        		      
        		// try to kick to score if we are near to opponent goal
      			if (posAgent.getDistanceTo(goalMiddle) <= (PITCH_LENGTH / 3.0))
            	{
                	Log.log(100, "kopem...");
    				soc = kickToScore();
            	}
                // or try to pass to teammate
      			else
            	{
                	Log.log(100, "prihravam...");
                    double fit = 0;
      				soc = kickToPass(&fit);
                    
                    // if fitness is too bad then kick to score
                    if (fit < 5 && posAgent.getDistanceTo(goalMiddle) >= (PITCH_LENGTH / 4))
                        soc = kickToScore();
            	}
                ACT->putCommandInQueue( soc );
                ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		}
    	}
    	else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    	{                                                // if fastest to ball
      		Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      		soc = intercept( false );                      // intercept the ball

      		if( soc.commandType == CMD_DASH &&             // if stamina low
          			WM->getAgentStamina().getStamina() <
             		SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      		{
        		soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        		//ACT->putCommandInQueue( soc );
        		//ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		}
      		else                                           // if stamina high
      		{
        		//ACT->putCommandInQueue( soc );               // dash as intended
        		//ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      		}
      		
      		
      		// DOPLNENE
      		ACT->putCommandInQueue( soc );
  			oscillateView(soc);	// kmitanie hlavickou ked bezim k lopte;    
      		// DOPLNENE KONIEC
     	}
     	else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     	{
       		if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+800 )
       		{
         		soc = moveToPos(WM->getStrategicPosition(),
                         PS->getPlayerWhenToTurnAngle());
        		ACT->putCommandInQueue( soc );            // move to strategic pos
         		//ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
         		oscillateView(soc); 
       		}
       		else                                        // else watch ball
       		{
         		ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         		//ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
         		oscillateView(soc); 
       		}
     	}
     	else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     	{
       		ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       		ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     	}
     	else                                         // nothing to do
       		ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   	}
  	return soc;
}

/*!This method is a simple goalie based on the goalie of the simple Team of
   FC Portugal. It defines a rectangle in its penalty area and moves to the
   position on this rectangle where the ball intersects if you make a line
   between the ball position and the center of the goal. If the ball can
   be intercepted in the own penalty area the ball is intercepted and catched.
*/
SoccerCommand Player::deMeer5_goalie(  )
{
  int i;
  SoccerCommand soc;
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle();

  // define the top and bottom position of a rectangle in which keeper moves
  static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
  static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );

  // define the borders of this rectangle using the two points.
  static Line  lineFront = Line::makeLineFromTwoPoints(posLeftTop,posRightTop);
  static Line  lineLeft  = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posLeftTop.getY()), posLeftTop );
  static Line  lineRight = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posRightTop.getY()),posRightTop );

  // nastavenie pohladu
  double distance = posAgent.getDistanceTo( WM->getBallPos() );
    
  // nastavenie pohladu
  SoccerCommand sc = adjustView( distance , 20, 50);
  if(sc.commandType != CMD_ILLEGAL) {
 	ACT->putCommandInQueue(sc);
  }

  if( WM->isBeforeKickOff( ) )
  {
    if( formations->getFormation() != FT_INITIAL || // not in kickoff formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )  
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( soc=teleportToPos(WM->getStrategicPosition()) );
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( soc = turnBodyToPoint( VecPosition( 0, 0 ), 0 ));
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
    return soc;
  }

  if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
  {                                                // confidence ball too  low
    ACT->putCommandInQueue( searchBall() );        // search ball
    ACT->putCommandInQueue( alignNeckWithBody( ) );
  }
  else if( WM->getPlayMode() == PM_PLAY_ON || WM->isFreeKickThem() ||
           WM->isCornerKickThem() )               
  {
    if( WM->isBallCatchable() )
    {
      ACT->putCommandInQueue( soc = catchBall() );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
     else if( WM->isBallKickable() )
    {
       //soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );    
       soc = kickToPass( );       
       ACT->putCommandInQueue( soc );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else if( WM->isInOwnPenaltyArea( getInterceptionPointBall( &i, true ) ) &&
             WM->getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &i ) == 
                                               WM->getAgentObjectType() )
    {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
    {
      //lineBack = Line::makeLineFromTwoPoint(VecPosition( -50.0, posLeftTop.getY()), VecPosition( -50.0, posRightTop.getY()));
      
      // vrhni sa po lopte ak je dostatocne blizko
      if(
         ((WM->getAgentGlobalPosition() - WM->getGlobalPosition( OBJECT_BALL )).getMagnitude() < 8 )
         //(WM->getBallSpeed() < 1.5f)
      )
      {          
          int cyc;          
          ObjectT oponent = WM->getFastestInSetTo( OBJECT_SET_OPPONENTS, OBJECT_BALL, &cyc);
          
          if((oponent != OBJECT_ILLEGAL) && (cyc <= 1))
          {
              ACT->putCommandInQueue( dashToPoint( WM->getGlobalPosition( oponent ), 1 ));
          }
      }
      else {
	      /*
	      Line lineBall1 = Line::makeLineFromPositionAndAngle(WM->getBallPos(), WM->getBallDirection());
	      
	      VecPosition goalWhere = lineBall1.getIntersection(lineBack);
	      
	      if((abs(goalWhere.getY()) < (SS->getGoalWidth() / 2)) && (goalWhere.getX() == -50))
	      {
	         // lopta smeruje do branky
	         VecPosition inter = lineBall1.getPointOnLineClosestTo(WM->getAgentGlobalPosition());
	         ACT->putCommandInQueue( turnBodyToPoint(inter) );
	         ACT->
	      }*/
	      
	      
	                 
	
	      // make line between own goal and the ball
	      VecPosition posMyGoal = ( WM->getSide() == SIDE_LEFT )
	             ? SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_L, SIDE_LEFT )
	             : SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_R, SIDE_RIGHT);
	      Line lineBall = Line::makeLineFromTwoPoints( WM->getBallPos(),posMyGoal);
	
	      // determine where your front line intersects with the line from ball
	      VecPosition posIntersect = lineFront.getIntersection( lineBall );
	
	      // outside rectangle, use line at side to get intersection
	      if (posIntersect.isRightOf( posRightTop ) )
	        posIntersect = lineRight.getIntersection( lineBall );
	      else if (posIntersect.isLeftOf( posLeftTop )  )
	        posIntersect = lineLeft.getIntersection( lineBall );
	
	      if( posIntersect.getX() < -49.0 )
	        posIntersect.setX( -49.0 );
	        
	      // and move to this position
	      if( posIntersect.getDistanceTo( WM->getAgentGlobalPosition() ) > 0.5 )
	      {
	        soc = moveToPos( posIntersect, PS->getPlayerWhenToTurnAngle() );
	        ACT->putCommandInQueue( soc );
	        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
	      }
	      else
	      {
	        ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
	        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
	      }
	    }
      }
  }
  else if( WM->isFreeKickUs() == true || WM->isGoalKickUs() == true )
  {
    if( WM->isBallKickable() )
    {
      if( WM->getTimeSinceLastCatch() == 25 && WM->isFreeKickUs() )
      {
        // move to position with lesser opponents.
        if( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, 
                                          Circle(posRightTop, 15.0 )) <
            WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, 
                                           Circle(posLeftTop,  15.0 )) )
          soc.makeCommand( CMD_MOVE,posRightTop.getX(),posRightTop.getY(),0.0);
        else
          soc.makeCommand( CMD_MOVE,posLeftTop.getX(), posLeftTop.getY(), 0.0);
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() > 28 )
      {
        //soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );    
        soc = kickToPass( );    
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() < 25 )
      {
        VecPosition posSide( 0.0, posAgent.getY() ); 
        if( fabs( (posSide - posAgent).getDirection() - angBody) > 10 )
        {
          soc = turnBodyToPoint( posSide );
          ACT->putCommandInQueue( soc );
        }
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isGoalKickUs()  )
    {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  else
  {
     ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
     ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  return soc;
}
