#include "fuzzyobj.h"
#include <stdio.h>
//#include <unistd.h>
#include <stdlib.h>

cFuzzyObj::cFuzzyObj()
{
	input=NULL,output=NULL,rule=NULL;
	inputs=outputs=rules=0;
}
cFuzzyObj::~cFuzzyObj()
{
	int i;

	for(i=0;i<inputs;i++) delete(input[i]);
	for(i=0;i<outputs;i++) delete(output[i]);
	for(i=0;i<rules;i++) delete(rule[i]);
	if(input!=NULL) free(input);
	if(output!=NULL) free(output);
	if(rule!=NULL) free(rule);
}
void cFuzzyObj::SetSamples(int samples1)
{
	samples=samples1;
}
void cFuzzyObj::AddInput(float min1,float max1)
{
	inputs++;
	input=(fuzzygroup **)realloc(input,inputs*sizeof(fuzzygroup *));
	input[inputs-1]=new fuzzygroup(min1,max1,samples);
}
void cFuzzyObj::AddOutput(float min1,float max1)
{
	outputs++;
	output=(fuzzygroup **)realloc(output,outputs*sizeof(fuzzygroup *));
	output[outputs-1]=new fuzzygroup(min1,max1,samples);
}
void cFuzzyObj::AddInputMF(int inputgroup1,...)
{
	va_list list;
	double arg1,arg2,arg3,arg4;
	int typ;
		
	va_start(list,inputgroup1);
	typ=va_arg(list,int);
	
	input[inputgroup1]->addfuzzy();

	if(typ==FZ_TRIANGLE)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		arg3=va_arg(list,double);
		input[inputgroup1]->mf[input[inputgroup1]->mfs-1]->settriangle(arg1,arg2,arg3);
	}
	if(typ==FZ_TRAPEZOID)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		arg3=va_arg(list,double);
		arg4=va_arg(list,double);
		input[inputgroup1]->mf[input[inputgroup1]->mfs-1]->settrapezoid(arg1,arg2,arg3,arg4);
	}
	if(typ==FZ_LEFT)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		input[inputgroup1]->mf[input[inputgroup1]->mfs-1]->setleft(arg1,arg2);
	}
	if(typ==FZ_RIGHT)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		input[inputgroup1]->mf[input[inputgroup1]->mfs-1]->setright(arg1,arg2);
	}
	if(typ==FZ_CONST)
	{
		arg1=va_arg(list,double);
		input[inputgroup1]->mf[input[inputgroup1]->mfs-1]->setconst(arg1);
	}

	if(typ==FZ_WELL) {
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		arg3=va_arg(list,double);
		arg4=va_arg(list,double);
		input[inputgroup1]->mf[input[inputgroup1]->mfs-1]->setwell(arg1,arg2,arg3,arg4);
	}
}
void cFuzzyObj::AddOutputMF(int outputgroup1,...)
{
	va_list list;
	double arg1,arg2,arg3,arg4;
	int typ;
		
	va_start(list,outputgroup1);
	typ=va_arg(list,int);
	
	output[outputgroup1]->addfuzzy();
	if(typ==FZ_TRIANGLE)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		arg3=va_arg(list,double);
		output[outputgroup1]->mf[output[outputgroup1]->mfs-1]->settriangle(arg1,arg2,arg3);
	}
	if(typ==FZ_TRAPEZOID)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		arg3=va_arg(list,double);
		arg4=va_arg(list,double);
		output[outputgroup1]->mf[output[outputgroup1]->mfs-1]->settrapezoid(arg1,arg2,arg3,arg4);
	}
	if(typ==FZ_LEFT)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		output[outputgroup1]->mf[output[outputgroup1]->mfs-1]->setleft(arg1,arg2);
	}
	if(typ==FZ_RIGHT)
	{
		arg1=va_arg(list,double);
		arg2=va_arg(list,double);
		output[outputgroup1]->mf[output[outputgroup1]->mfs-1]->setright(arg1,arg2);
	}
	if(typ==FZ_CONST)
	{
		arg1=va_arg(list,double);
		output[outputgroup1]->mf[output[outputgroup1]->mfs-1]->setconst(arg1);
	}
	va_end(list);
}
void cFuzzyObj::AddRule(int outputgroup1,...)
{
	va_list list;
	int i;

	va_start(list,outputgroup1);
	
	rules++;
	rule=(fuzzyrule **)realloc(rule,rules*sizeof(fuzzyrule *));
	rule[rules-1]=new fuzzyrule(inputs);

	rule[rules-1]->setoutputgroup(outputgroup1);
	for(i=0;i<inputs;i++) {
		rule[rules-1]->setinput(i,va_arg(list,int));
	}
	rule[rules-1]->setoutput(va_arg(list,int));
	va_end(list);
}
float cFuzzyObj::CountResult(int outputgroup1,int method,int ops,...)
{
	int i,j,k;
	float val;
	va_list list;
	double *values;
	
	values=(double *)malloc(inputs*sizeof(double));
	va_start(list,ops);
	
	for(i=0;i<inputs;i++) values[i]=va_arg(list,double);
		
	fuzzy out(output[outputgroup1]->min,output[outputgroup1]->max,output[outputgroup1]->samples);
	fuzzy c(output[outputgroup1]->min,output[outputgroup1]->max,output[outputgroup1]->samples);
	out.setconst(0);

	for(i=0;i<rules;i++) {
		if(rule[i]->outputgroup==outputgroup1)
		{
			val=1;
			for(j=0;j<rule[i]->inputs;j++)
			{
				if(rule[i]->input[j]>=0)
				{
					if(ops==OPS_ZADEH) val=andz(val,input[j]->mf[rule[i]->input[j]]->value(values[j]));
					if(ops==OPS_LARSEN) val=andp(val,input[j]->mf[rule[i]->input[j]]->value(values[j]));
					if(ops==OPS_LUKASIEWITZ) val=andl(val,input[j]->mf[rule[i]->input[j]]->value(values[j]));
				}
			}
			if(val>0)
			{
				if(ops==OPS_ZADEH) c.setandz(*output[outputgroup1]->mf[rule[i]->output],val);
				if(ops==OPS_LARSEN)	c.setandp(*output[outputgroup1]->mf[rule[i]->output],val);
				if(ops==OPS_LUKASIEWITZ) c.setandl(*output[outputgroup1]->mf[rule[i]->output],val);

				out.setorz(out,c);
			}
		}
	}
	if(values!=NULL) {
		free(values);
	}
	va_end(list);

	if(method==DFZ_CENTROID) return(out.centroid());
	if(method==DFZ_BISECTOR) return(out.bisector());
	if(method==DFZ_LOM) return(out.lom());
	if(method==DFZ_MOM) return(out.mom());
	if(method==DFZ_HOM) return(out.hom());
	return(out.centroid());
} 
